#include <iostream>
#include <type_traits>
#include <sys/types.h>
#include <unistd.h>
#include <cxxtest/TestSuite.h>

#include <async_stream.hpp>
#include <file_stream.hpp>
#include <block_stream.hpp>
#include <buffer_stream.hpp>
#include <virtual_stream.hpp>

#include <iostream.hpp>

template<class T, class POD,
    typename
    std::enable_if<
        std::is_same<
            std::streamsize,
            decltype(((T*)0x0)->read(nullptr, std::streamsize(0x0)))
        >::value
    >::type* = nullptr
>
T &operator>>(T &stream, POD &output) {
    stream.read(&output, sizeof(POD));
    return stream;
}

std::string workingDir() {
    char szTmp[32];
    constexpr int len=2028;
    char pBuf[len];

    sprintf(szTmp, "/proc/%d/exe", getpid());
    int bytes = std::min<int>(readlink(szTmp, pBuf, len), len - 1);
    if(bytes >= 0)
        pBuf[bytes] = '\0';
    return pBuf;
}
class MyTestSuite : public CxxTest::TestSuite
{
public:
    void testAddition( void )
    {
        TS_ASSERT( 1 + 1 > 1 );
        TS_ASSERT_EQUALS( 1 + 1, 2 );
    }

    int countLines(uint8_t* it, uint8_t* end) {
        int lines = 0;
        while(it != end) {
            if(*it == '\n') {
                ++lines;
            }
            ++it;
        }
        return lines;
    }

    void testMemoryStream() {
        MemoryDuplexStream stream(1024);
        stream.write(100);
        TS_ASSERT_EQUALS(stream.tell(), 100);
        TS_ASSERT_EQUALS(stream.size(), 100);
        TS_ASSERT(stream.capacity() == 1024);
    }

    template<typename T>
    void simpleInputStreamTest(Data<> &expectedOutput, T &stream) {
        std::vector<uint8_t> loadedFileData;
        int linesRead = 0;
        for(std::string line : lineReader(std::move(stream))) {
            // so ineffeciant
            for(std::size_t i = 0; i < line.size(); ++i) {
                loadedFileData.push_back(line[i]);
            }
            loadedFileData.push_back('\n');
            ++linesRead;
        }

        Data<> data(new uint8_t[loadedFileData.size()], loadedFileData.size());
        std::copy(loadedFileData.begin(), loadedFileData.end(), data.begin());

        TS_ASSERT(expectedOutput == data);
        int lines = countLines(expectedOutput.begin(), expectedOutput.end());
        TS_ASSERT_EQUALS(lines, linesRead);
    }
    void testAsync() {
        Data<> fileData = loadFile("testFile.txt");

        file_istream inputFile("testFile.txt");
        TS_ASSERT(inputFile.is_open());
        async_istream<file_istream> inputAsync(std::move(inputFile));
        simpleInputStreamTest(fileData, inputAsync);
    }

    void testBlockInputStream() {
        Data<> fileData = loadFile("testFile.txt");

        file_istream inputFile("testFile.txt");
        TS_ASSERT(inputFile.is_open());
        block_istream<file_istream> input(std::move(inputFile), 512);
        simpleInputStreamTest(fileData, input);
    }

    void testBlockInputStream111() {
        // this is to test a weird block size
        Data<> fileData = loadFile("testFile.txt");

        file_istream inputFile("testFile.txt");
        TS_ASSERT(inputFile.is_open());
        block_istream<file_istream> input(std::move(inputFile), 111);
        simpleInputStreamTest(fileData, input);
    }

    void testBufferInputStream() {
        // this is to test a weird block size
        Data<> fileData = loadFile("testFile.txt");

        file_istream inputFile("testFile.txt");
        TS_ASSERT(inputFile.is_open());
        buffer_istream<file_istream> input(std::move(inputFile));
        simpleInputStreamTest(fileData, input);
    }

    void testChainInputStream() {
        // this is to test a weird block size
        Data<> fileData = loadFile("testFile.txt");

        file_istream inputFile("testFile.txt");
        TS_ASSERT(inputFile.is_open());
        chain_istream input(std::move(inputFile));
        input.push_new<buffer_istream<chain_istream::ChainStream>>();

        simpleInputStreamTest(fileData, input);
    }

    void testCopy() {
        // this is to test a weird block size
        Data<> fileData = loadFile("testFile.txt");

        file_istream inputFile("testFile.txt");
        TS_ASSERT(inputFile.is_open());
        buffer_istream<file_istream> input(std::move(inputFile));
        heap_iostream<> output;
        copy_result result = copy(input, output);
        TS_ASSERT_EQUALS(result.total_write, result.total_read);
        TS_ASSERT(result.success);

        TS_ASSERT_EQUALS(output.tell(), result.total_write);
        TS_ASSERT_EQUALS(output.tell(), output.size());
        TS_ASSERT_EQUALS(output.tell(), fileData.size());
        std::streamsize size = output.tell();
        output.seek(0);
        Data<> copied(new uint8_t[size], size);
        std::streamsize data_read = output.read(copied.get(), copied.size());
        TS_ASSERT_EQUALS(data_read, copied.size());
        TS_ASSERT_EQUALS(copied.size(), size);
        TS_ASSERT(copied == fileData);
    }
};

